#include <stdio.h>
#include "rawfmt.h"

/********* SUN RASTER FILES ***************/
/*
 * Description of header for files containing raster images
 */
struct rasterfile {
	long     ras_magic;              /* magic number */
	long     ras_width;              /* width (pixels) of image */
	long     ras_height;             /* height (pixels) of image */
	long     ras_depth;              /* depth (1, 8, or 24 bits) of pixel */
	long     ras_length;             /* length (bytes) of image */
	long     ras_type;               /* type of file; see RT_* below */
	long     ras_maptype;            /* type of colormap; see RMT_* below */
	long     ras_maplength;          /* length (bytes) of following map */
	/* color map follows for ras_maplength bytes, followed by image */
};
#define RAS_MAGIC       0x59a66a95

        /* Sun supported ras_type's */
#define RT_OLD          0       /* Raw pixrect image in 68000 byte order */
#define RT_STANDARD     1       /* Raw pixrect image in 68000 byte order */
#define RT_BYTE_ENCODED 2       /* Run-length compression of bytes */
#define RT_EXPERIMENTAL 0xffff  /* Reserved for testing */

        /* Sun registered ras_maptype's */
#define RMT_RAW         2
        /* Sun supported ras_maptype's */
#define RMT_NONE        0       /* ras_maplength is expected to be 0 */
#define RMT_EQUAL_RGB   1       /* red[ras_maplength/3],green[],blue[] */


StringPtr rastName;

rasterr(char *fmt, long arg)
{
	PtoCstr((char *)rastName);
	fprintf(stderr,"  %s: SunRast: ", rastName);
	fprintf(stderr,fmt, arg);
	CtoPstr((char *)rastName);
}


/* SUN raster file */
GetSunRast(HFileInfo *hp, raw_t *img)
{
	struct rasterfile rh;
	register unsigned char *cp, *fileData = NULL;
	long n, len, filesize;
	int refNum, i,j;
	register int c;
	int encoded=0;
	int err, rch, rcnt=0;
	int rc = FALSE;
	
	rastName = hp->ioNamePtr;
	
	/* open file 'Mac' style */
	if ((err=HOpen(hp->ioVRefNum, hp->ioDirID, hp->ioNamePtr, fsRdPerm, &refNum)) != noErr) {
		rasterr("HOpen err = %ld\n",err);
		goto done;
	}
		
	/* read and validate header */
	n = sizeof(rh);
	if ((err=FSRead(refNum, &n, &rh)) != noErr)
		goto done;

	if (rh.ras_magic != RAS_MAGIC)
		goto done;

	if (rh.ras_type == RT_BYTE_ENCODED)
		encoded = 1;
	else if (rh.ras_type != RT_STANDARD) {
		rasterr("Don't know how to read type %ld rasterfile\n", rh.ras_type);
		goto done;
	}
	if (rh.ras_depth > 8) {
		rasterr("Don't know how to read %ld depth rasterfiles\n", rh.ras_depth);
		goto done;
	}
	
	img->width = rh.ras_width;
	img->height = rh.ras_height;

	len = rh.ras_width * rh.ras_height;
	img->data = (unsigned char *)NewPtr(len);
	if (img->data == 0L) {
		rasterr("out of mem (raster)\n",0);
		goto done;
	}

	if (rh.ras_maplength) {
		if (rh.ras_maptype != RMT_EQUAL_RGB) {
			rasterr("Don't know how to read rasterfile map type %ld", rh.ras_maptype);
			goto done;
		}
		/* borrow big buf for reading in color table */
		n = rh.ras_maplength;
		if ((err=FSRead(refNum, &n, img->data)) != noErr)
			goto done;
		
		/* build color table */
		n /= 3;		/* r,g,b */
		cp = img->data;
		for (i=0; i<n; i++)
			img->cmap[i][RED] = cp[i];
		cp += n;
		for (i=0; i<n; i++)
			img->cmap[i][GREEN] = cp[i];
		cp += n;
		for (i=0; i<n; i++)
			img->cmap[i][BLUE] = cp[i];
		
		img->ncolors = n;
	}
	/* if 8bit, unencoded, the data is already in the correct format */
	if (rh.ras_depth == 8 && !encoded) {
		n = len;
		if ((err=FSRead(refNum, &n, img->data)) == noErr)
			rc = TRUE;
		goto done;
	}
	
	/* have to do it the hard way -- be a memory hog, read in entire file */
	err=GetEOF(refNum, &filesize);
	n = filesize;
	if ((fileData = (uchar *)NewPtr(n)) == NULL) {
		rasterr("out of mem (file data)\n",0);
		goto done;
	}
	if ((err=FSRead(refNum, &n, fileData)) != noErr) {
		DisposPtr(fileData);
		goto done;
	}
	
	if (rh.ras_depth == 8 && encoded) {
		cp = img->data;
		while (len--) {
			if (rcnt) {
				rcnt--;
				c = rch;
			}
			else if ((c = *fileData++) == 0x80) {	/* rle escape */
				if ((rcnt = *fileData++) == 0)		/* count */
					c = 0x80;
				else c = rch = *fileData++;
			}
			*cp++ = c;
		}
	}
	else if (rh.ras_depth == 1) {
		/* make a grey scale table */
		for (i=0; i<256; i++)
			img->cmap[i][RED] = img->cmap[i][GREEN] = img->cmap[i][BLUE] = i;
		img->ncolors = 256;

		/* convert bits to bytes */
		cp = img->data;
		len /= 8;
		while (len--) {
			if (rcnt) {
				rcnt--;
				c = rch;
			}
			else if ((c = *fileData++) == 0x80 && encoded) {	/* rle escape */
				if ((rcnt = *fileData++) == 0)		/* count */
					c = 0x80;
				else c = rch = *fileData++;
			}
			for (i=0,j=0x80; i<8; i++,j>>=1)
				*cp++ = (j & c) ? 0 : 255;
		}
	}
	rc = TRUE;
done:
	if (fileData)
		DisposPtr(fileData);
	if (err && err != eofErr)
		rasterr("I/O error = %ld\n",err);
	FSClose(refNum);
	return (rc);
}
