/*
 * convert TIFF files to 'raw' color map
 *
 * NOTE: this requires:
 * 	TIFF Software, by Sam Leffler <sam@okeeffe.berkeley.edu>.  Nice
 * 	portable library for reading and writing TIFF files, plus a few tools
 * 	for manipulating them and reading other formats.  Available via FTP as
 * 	okeeffe.berkeley.edu:pub/tiff.tar.Z.
 */
#include <stdio.h>
#include "tiffio.h"
#include "rawfmt.h"

char       *inf;

tifferr(char *fmt, int arg)
{
	fprintf(stderr,"%s: TIFF: ", inf);
	fprintf(stderr,fmt, arg);
}

TIFFToRaw(HFileInfo *hp, raw_t *img)
{
    short       width,
                height;
    int         depth,
                numcolors;
    register TIFF *tif = NULL;
    TIFFDirectory *td;
    register u_char *inp, *outp;
    register int x, y, i;
    u_char     *buf = NULL;
	WDParam wd;
	int err;
	int rc = FALSE;
                
    /* input file name */
	PtoCstr((char *)hp->ioNamePtr);
    inf = (char *)hp->ioNamePtr;

    /*         
     * !!! rather than rewrite the TIFF library to use Mac calls, setup a working directory
     *     and 'cd' to it.
     */
	wd.ioNamePtr = NULL;
	wd.ioVRefNum = hp->ioVRefNum;
	wd.ioWDProcID = 0L;
	wd.ioWDDirID = hp->ioDirID;
	if ((err = PBOpenWD(&wd, FALSE)) != noErr) {
		tifferr("PBOpenWD err = %d\n",err);
		goto done;
	}
	SetVol(0L, wd.ioVRefNum);

    tif = TIFFOpen(inf, "r");
    if (tif == NULL)
		goto done;

    td = &tif->tif_dir;

    if (td->td_bitspersample > 8) {
		tifferr("can't handle more than 8-bits per sample\n", 0);
		goto done;
	}

    switch (td->td_samplesperpixel) {
    case 1:
		if (td->td_bitspersample == 1)
		    depth = 1;
		else
		    depth = 8;
		break;
    case 3:
    case 4:
		depth = 24;
		break;
    default:
		tifferr("only handle 1-channel gray scale or 3-channel color\n", 0);
		goto done;
    }

    width = td->td_imagewidth;
    height = td->td_imagelength;

	img->width = width;
	img->height = height;
	img->data = (uchar *)NewPtr((long)width * (long)height);
	if (img->data == NULL) {
		tifferr("out of memory (image data)\n", 0);
		goto done;
	}

    numcolors = (1 << td->td_bitspersample);
	img->ncolors = numcolors;

    if (numcolors == 2) {	/* monochrome */
		img->cmap[0][RED] = img->cmap[0][GREEN] = img->cmap[0][BLUE] = 0xFF;
		img->cmap[1][RED] = img->cmap[1][GREEN] = img->cmap[1][BLUE] = 0x00;
    } else
		switch (td->td_photometric) {
		case PHOTOMETRIC_MINISBLACK:
		    for (x = 0; x < numcolors; x++)
				img->cmap[x][RED] = img->cmap[x][GREEN] = img->cmap[x][BLUE] = 
						(255 * x) / numcolors;
		    break;
		case PHOTOMETRIC_RGB:		/* too lazy to find best colors */
		case PHOTOMETRIC_MINISWHITE:
		    for (x = 0; x < numcolors; x++)
				img->cmap[x][RED] = img->cmap[x][GREEN] = img->cmap[x][BLUE] = 
						255 - ((255 * x) / numcolors);
		    break;
		case PHOTOMETRIC_PALETTE:
		    for (i = 0; i < numcolors; i++) {
#define	CVT(x)		(((x) >> 8))
				img->cmap[i][RED] = (u_char) CVT(td->td_redcolormap[i]);
				img->cmap[i][GREEN] = (u_char) CVT(td->td_greencolormap[i]);
				img->cmap[i][BLUE] = (u_char) CVT(td->td_bluecolormap[i]);
		    }
		    break;
		case PHOTOMETRIC_MASK:
		    tifferr("Don't know how to handle PHOTOMETRIC_MASK\n", 0);
		    goto done;
		case PHOTOMETRIC_DEPTH:
		    tifferr("Don't know how to handle PHOTOMETRIC_DEPTH\n", 0);
		    goto done;
		default:
		    tifferr("unknown photometric (cmap): %d\n", td->td_photometric);
		    goto done;
		}

    buf = (u_char *) NewPtr((long)width * (long)td->td_samplesperpixel);
    if (buf == NULL) {
		tifferr("can't allocate memory for scanline buffer...\n", 0);
		goto done;
	}

    for (y = 0; (y < height) && (TIFFReadScanline(tif, buf,y,0) > 0); y++) {
		inp = buf;
		outp = (u_char *) &img->data[(long)y*(long)width];

		switch (td->td_photometric) {
		case PHOTOMETRIC_RGB:
#define	MkMono(r,g,b) ( ((long)(r) * 35L)/100L + ((long)(g) * 53L)/100L + ((long)(b) * 12L)/100L)

		    if (td->td_samplesperpixel == 4)
			for (x = 0; x < width; x++) {
				*outp++ = MkMono(inp[0],inp[1],inp[2]);
				inp += 4;	/* skip alpha channel */
			}
		    else
			for (x = 0; x < width; x++) {
				*outp++ = MkMono(inp[0],inp[1],inp[2]);
				inp += 3;
			}
		    break;
		case PHOTOMETRIC_MINISWHITE:
		case PHOTOMETRIC_MINISBLACK:
		    switch (td->td_bitspersample) {
		    case 1:
				for (x = 0; x < ((width + 7) / 8); x++) {
				    *outp++ = (*inp & 0x80) ? 1 : 0;
				    *outp++ = (*inp & 0x40) ? 1 : 0;
				    *outp++ = (*inp & 0x20) ? 1 : 0;
				    *outp++ = (*inp & 0x10) ? 1 : 0;
				    *outp++ = (*inp & 0x08) ? 1 : 0;
				    *outp++ = (*inp & 0x04) ? 1 : 0;
				    *outp++ = (*inp & 0x02) ? 1 : 0;
				    *outp++ = (*inp & 0x01) ? 1 : 0;
				    inp++;
				}
				break;
		    case 2:
				for (x = 0; x < ((width + 3) / 4); x++) {
				    *outp++ = (*inp >> 6) & 3;
				    *outp++ = (*inp >> 4) & 3;
				    *outp++ = (*inp >> 2) & 3;
				    *outp++ = *inp++ & 3;
				}
				break;
		    case 4:
				for (x = 0; x < width / 2; x++) {
				    *outp++ = *inp >> 4;
				    *outp++ = *inp++ & 0xf;
				}
				break;
		    case 8:
				for (x = 0; x < width; x++)
				    *outp++ = *inp++;
				break;
		    default:
				tifferr("bad bits/sample: %d\n", td->td_bitspersample);
				goto done;
		    }
		    break;
		case PHOTOMETRIC_PALETTE:
		    for (x = 0; x < width; x++)
			*outp++ = *inp++;
		    break;
		default:
		   tifferr("unknown photometric (write): %d\n", td->td_photometric);
		   goto done;
		}
    }
    if (y < height) {
		tifferr("bad data read on line: %d\n", y);
		goto done;
	}

	rc = TRUE;
done:
	CtoPstr((char *)hp->ioNamePtr);
    PBCloseWD(&wd, FALSE);
	if (buf)
    	DisposPtr(buf);
    if (tif)
    	TIFFClose(tif);
    
    return (rc);
}
